""" 
MAIN MODULE FOR PLANEAT

ANAPLAN document: https://github.com/HCB06/Anaplan/blob/main/Welcome_to_Anaplan/ANAPLAN_USER_MANUEL_AND_LEGAL_INFORMATION(EN).pdf

@author: Hasan Can Beydili
@YouTube: https://www.youtube.com/@HasanCanBeydili
@Linkedin: https://www.linkedin.com/in/hasan-can-beydili-77a1b9270/
@Instagram: https://www.instagram.com/canbeydili.06/
@contact: tchasancan@gmail.com
"""

import numpy as np
import random
from tqdm import tqdm

### LIBRARY IMPORTS ###
from .plan import feed_forward
from .data_operations import normalization
from .ui import loading_bars, initialize_loading_bar
from .activation_functions import apply_activation, all_activations

def define_genomes(input_shape, output_shape, population_size, dtype=np.float32):
   """
   Initializes a population of genomes, where each genome is represented by a set of weights 
   and an associated activation function. Each genome is created with random weights and activation 
   functions are applied and normalized. (Max abs normalization.)

   Args:

      input_shape (int): The number of input features for the neural network.

      output_shape (int): The number of output features for the neural network.
      
      population_size (int): The number of genomes (individuals) in the population.

      dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

   Returns:
      tuple: A tuple containing:
         - population_weights (numpy.ndarray): A 2D numpy array of shape (population_size, output_shape, input_shape) representing the 
            weight matrices for each genome.
         - population_activations (list): A list of activation functions applied to each genome.
            
   Notes:
      The weights are initialized randomly within the range [-1, 1]. 
      Activation functions are selected randomly from a predefined list `all_activations()`.
      The weights for each genome are then modified by applying the corresponding activation function 
      and normalized using the `normalization()` function. (Max abs normalization.)
   """
   population_weights = [0] * population_size
   population_activations = [0] * population_size

   except_this = ['spiral', 'circular']
   activations = [item for item in all_activations() if item not in except_this] # SPIRAL AND CIRCULAR ACTIVATION DISCARDED

   for i in range(len(population_weights)):

      population_weights[i] = np.random.uniform(-1, 1, (output_shape, input_shape)).astype(dtype)
      population_activations[i] = activations[int(random.uniform(0, len(activations)-1))]

      # ACTIVATIONS APPLYING IN WEIGHTS SPECIFIC OUTPUT CONNECTIONS (MORE PLAN LIKE FEATURES(FOR NON-LINEARITY)):

      for j in range(population_weights[i].shape[0]):

         population_weights[i][j,:] = apply_activation(population_weights[i][j,:], population_activations[i])
         population_weights[i][j,:] = normalization(population_weights[i][j,:], dtype=dtype)

   return np.array(population_weights, dtype=dtype), population_activations


def evolve(weights, activation_potentiations, what_gen, fitness, show_info=False, strategy='cross_over', bar_status=True, policy='normal_selective', target_fitness='max', mutations=True, bad_genoms_mutation_prob=None, activation_mutate_prob=0.5, save_best_genom=True, cross_over_mode='tpm', activation_add_prob=0.5, activation_delete_prob=0.5, activation_change_prob=0.5, weight_mutate_prob=1, weight_mutate_rate=32, activation_selection_add_prob=0.7, activation_selection_change_prob=0.5, activation_selection_rate=2, dtype=np.float32):
   """
Applies the evolving process of a population of genomes using selection, crossover, mutation, and activation function potentiation.
The function modifies the population's weights and activation functions based on a specified policy, mutation probabilities, and strategy.

Args:
    weights (numpy.ndarray): Array of weights for each genome. 
        (first returned value of define_genomes function)
    
    activation_potentiations (list): A list of activation functions for each genome. 
        (second returned value of define_genomes function)
    
    what_gen (int): The current generation number, used for informational purposes or logging.
    
    fitness (numpy.ndarray): A 1D array containing the fitness values of each genome. 
        The array is used to rank the genomes based on their performance. PLANEAT maximizes or minimizes this fitness for looking 'target' hyperparameter.
    
    show_info (bool, optional): If True, prints information about the current generation and the 
        maximum reward obtained. Also shows the current configuration. Default is False.
    
    strategy (str, optional): The strategy for combining the best and bad genomes. Options:
        - 'cross_over': Perform crossover between the best genomes and replace bad genomes. 
          (Classic NEAT crossover)
        - 'potentiate': Cumulate the weight of the best genomes and replace bad genomes. 
          (PLAN feature, similar to arithmetic crossover but different.)
        Default is 'cross_over'.
    
    bar_status (bool, optional): Loading bar status during evolving process of genomes. True or False. Default: True

    policy (str, optional): The selection policy that governs how genomes are selected for reproduction. Options:
        - 'normal_selective': Normal selection based on reward, where a portion of the bad genes are discarded.
        - 'more_selective': A more selective policy, where fewer bad genes survive.
        - 'less_selective': A less selective policy, where more bad genes survive.
        Default is 'normal_selective'.
    
    target_fitness (str, optional): Target fitness strategy for PLANEAT optimization. ('max' for machine learning, 'min' for machine unlearning.) Default: 'max'     
   
    mutations (bool, optional): If True, mutations are applied to the bad genomes and potentially 
        to the best genomes as well. Default is True.
    
    bad_genoms_mutation_prob (float, optional): The probability of applying mutation to the bad genomes. 
        Must be in the range [0, 1]. Also affects the mutation probability of the best genomes inversely. 
        For example, a value of 0.7 for bad genomes implies 0.3 for best genomes. Default is None, 
        which means it is determined by the `policy` argument.
    
    activation_mutate_prob (float, optional): The probability of applying mutation to the activation functions. 
        Must be in the range [0, 1]. Default is 0.5 (50%).
    
    save_best_genom (bool, optional): If True, ensures that the best genomes are saved and not mutated 
        or altered during reproduction. Default is True.
    
    cross_over_mode (str, optional): Specifies the crossover method to use. Options:
        - 'tpm': Two-Point Matrix Crossover
        - 'plantic': plantic Crossover
        Default is 'tpm'.
    
    activation_add_prob (float, optional): The probability of adding a new activation function to the genome for mutation. 
        Must be in the range [0, 1]. Default is 0.5.
    
    activation_delete_prob (float, optional): The probability of deleting an existing activation function 
        from the genome for mutation. Must be in the range [0, 1]. Default is 0.5.
    
    activation_change_prob (float, optional): The probability of changing an activation function in the genome for mutation. 
        Must be in the range [0, 1]. Default is 0.5.
    
    weight_mutate_prob (float, optional): The probability of mutating a weight in the genome. 
        Must be in the range [0, 1]. Default is 1.
    
    weight_mutate_rate (int, optional): If the value you enter here is equal to the result of input layer * output layer, 
        only a single weight will be mutated during each mutation process. If the value you enter here is half 
        of the result of input layer * output layer, two weights in the weight matrix will be mutated. 
        WARNING: if you don't understand do NOT change this value. Default is 32.

    activation_selection_add_prob (float, optional): The probability of adding an existing activation function for cross over.
        from the genome. Must be in the range [0, 1]. Default is 0.7. (WARNING! More higher values make models more complex. For fast training rise this value.)
    
    activation_selection_change_prob (float, optional): The probability of changing an activation function in the genome for cross over. 
        Must be in the range [0, 1]. Default is 0.5.
    
    activation_selection_rate (int, optional): If the activation list of a good genome is smaller than the value entered here, only one activation will undergo a crossover operation. In other words, this parameter controls the model complexity. Default is 2.

    dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

Raises:
    ValueError: 
        - If `policy` is not one of the specified values ('normal_selective', 'more_selective', 'less_selective').
        - If `cross_over_mode` is not one of the specified values ('tpm', 'plantic').
        - If `bad_genoms_mutation_prob`, `activation_mutate_prob`, or other probability parameters are not in the range [0, 1].
        - If the population size is odd (ensuring an even number of genomes is required for proper selection).

Returns:
    tuple: A tuple containing:
        - weights (numpy.ndarray): The updated weights for the population after selection, crossover, and mutation. 
                                  The shape is (population_size, output_shape, input_shape).
        - activation_potentiations (list): The updated list of activation functions for the population.

Notes:
    - **Selection Process**: 
        - The genomes are sorted by their fitness (based on `fitness`), and then split into "best" and "bad" halves. 
        - The best genomes are retained, and the bad genomes are modified based on the selected strategy.
        
    - **Crossover and Potentiation Strategies**:
        - The **'cross_over'** strategy performs crossover, where parts of the best genomes' weights are combined with the other good genomes to create new weight matrices.
        - The **'potentiate'** strategy strengthens the best genomes by potentiating their weights towards the other good genomes.
        
    - **Mutation**:
        - Mutation is applied to both the best and bad genomes, depending on the mutation probability and the `policy`.
        - `bad_genoms_mutation_prob` determines the probability of applying mutations to the bad genomes.
        - If `activation_mutate_prob` is provided, activation function mutations are applied to the genomes based on this probability.
        
    - **Population Size**: The population size must be an even number to properly split the best and bad genomes. If `fitness` has an odd length, an error is raised.
    
    - **Logging**: If `show_info=True`, the current generation and the maximum reward from the population are printed for tracking the learning progress.

Example:
    ```python
    weights, activation_potentiations = planeat.evolve(weights, activation_potentiations, 1, fitness, show_info=True, strategy='cross_over', policy='normal_selective')
    ```

    - The function returns the updated weights and activations after processing based on the chosen strategy, policy, and mutation parameters.
"""
    
### ERROR AND CONFIGURATION CHECKS:

   if policy == 'normal_selective':
      if bad_genoms_mutation_prob == None:
         bad_genoms_mutation_prob = 0.7
      
   elif policy == 'more_selective':
      if bad_genoms_mutation_prob == None:
         bad_genoms_mutation_prob = 0.85

   elif policy == 'less_selective':
      if bad_genoms_mutation_prob == None:
         bad_genoms_mutation_prob = 0.6

   else:
      raise ValueError("policy parameter must be: 'normal_selective' or 'more_selective' or 'less_selective'")
   
   
   if (activation_add_prob < 0 or activation_add_prob > 1) or (activation_change_prob < 0 or activation_change_prob > 1) or (activation_delete_prob < 0 or activation_delete_prob > 1) or (weight_mutate_prob < 0 or weight_mutate_prob > 1) or (activation_selection_add_prob < 0 or activation_selection_add_prob > 1) or (activation_selection_change_prob < 0 or activation_selection_change_prob > 1):
      raise ValueError("All hyperparameters ending with 'prob' must be a number between 0 and 1.")

   if cross_over_mode != 'tpm' and cross_over_mode != 'plantic':
      raise ValueError("cross_over_mode parameter must be 'tpm' or 'plantic'")

   if bad_genoms_mutation_prob is not None:
      if not isinstance(bad_genoms_mutation_prob, float) or bad_genoms_mutation_prob < 0 or bad_genoms_mutation_prob > 1:
         raise ValueError("bad_genoms_mutation_prob parameter must be float and 0-1 range")
      
   if activation_mutate_prob is not None:
      if not isinstance(activation_mutate_prob, float) or activation_mutate_prob < 0 or activation_mutate_prob > 1:
         raise ValueError("activation_mutate_prob parameter must be float and 0-1 range")
      
   if len(fitness) % 2 == 0:
      slice_center = int(len(fitness) / 2)

   else:
      raise ValueError("genome population size must be even number. for example: not 99, make 100 or 98.")

   
### FITNESS IS SORTED IN ASCENDING (OR DESCENDING) ORDER, AND THE WEIGHT AND ACTIVATIONS OF EACH GENOME ARE SORTED ACCORDING TO THIS ORDER:

   if target_fitness == 'max': sort_indices = np.argsort(fitness)
   elif target_fitness == 'min': sort_indices = np.argsort(-fitness)

   fitness = fitness[sort_indices]
   weights = weights[sort_indices]

   activation_potentiations = [activation_potentiations[i] for i in sort_indices]

### GENOMES ARE DIVIDED INTO TWO GROUPS: GOOD GENOMES AND BAD GENOMES:

   best_weights = weights[slice_center:]
   bad_weights = weights[:slice_center]
   best_weight = best_weights[len(best_weights)-1]

   best_activations = list(activation_potentiations[slice_center:])
   bad_activations = list(activation_potentiations[:slice_center])
   best_activation = best_activations[len(best_activations) - 1]

   
### NEAT IS APPLIED ACCORDING TO THE SPECIFIED POLICY, STRATEGY, AND PROBABILITY CONFIGURATION:
   
   bar_format = loading_bars()[0]
   
   if bar_status: progress = initialize_loading_bar(len(bad_weights), desc="GENERATION: " + str(what_gen), bar_format=bar_format, ncols=50, ascii="▱▰")

   for i in range(len(bad_weights)):

      if policy == 'normal_selective':
            
         if strategy == 'cross_over':
            bad_weights[i], bad_activations[i] = cross_over(best_weight, best_weights[i], best_activations=best_activation, good_activations=best_activations[i], cross_over_mode=cross_over_mode, activation_selection_add_prob=activation_selection_add_prob, activation_selection_change_prob=activation_selection_change_prob, activation_selection_rate=activation_selection_rate)


         elif strategy == 'potentiate':
            bad_weights[i], bad_activations[i] = potentiate(best_weight, best_weights[i], best_activations=best_activation, good_activations=best_activations[i], dtype=dtype)
            
         
         if mutations is True:
            
            mutation_prob = random.uniform(0, 1)

            if mutation_prob > bad_genoms_mutation_prob:
               if (save_best_genom == True and not np.array_equal(best_weights[i], best_weight)) or save_best_genom == False:
                  best_weights[i], best_activations[i] = mutation(best_weights[i], best_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)
               
            elif mutation_prob < bad_genoms_mutation_prob:
               bad_weights[i], bad_activations[i] = mutation(bad_weights[i], bad_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)

      if policy == 'more_selective':
               
            if strategy == 'cross_over':            
               bad_weights[i], bad_activations[i] = cross_over(best_weight, best_weights[i], best_activations=best_activation, good_activations=best_activations[i], cross_over_mode=cross_over_mode, activation_selection_add_prob=activation_selection_add_prob, activation_selection_change_prob=activation_selection_change_prob, activation_selection_rate=activation_selection_rate)
            
            elif strategy == 'potentiate':
               bad_weights[i], bad_activations[i] = potentiate(best_weight, best_weights[i], best_activations=best_activation, good_activations=best_activations[i], dtype=dtype)
            
            if mutations is True:

               mutation_prob = random.uniform(0, 1)

               if mutation_prob > bad_genoms_mutation_prob:
                  if (save_best_genom == True and not np.array_equal(best_weights[i], best_weight)) or save_best_genom == False:
                     best_weights[i], best_activations[i] = mutation(best_weights[i], best_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)
                  
               elif mutation_prob < bad_genoms_mutation_prob:
                  bad_weights[i], bad_activations[i] = mutation(bad_weights[i], bad_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)



      if policy == 'less_selective':

            random_index = int(random.uniform(0, len(best_weights) - 1))
            
            if strategy == 'cross_over': 
               bad_weights[i], bad_activations[i] = cross_over(best_weights[random_index], best_weights[i], best_activations=best_activations[random_index], good_activations=best_activations[i], cross_over_mode=cross_over_mode, activation_selection_add_prob=activation_selection_add_prob, activation_selection_change_prob=activation_selection_change_prob, activation_selection_rate=activation_selection_rate)
            
            elif strategy == 'potentiate':
               bad_weights[i], bad_activations[i] = potentiate(best_weights[random_index], best_weights[i], best_activations=best_activations[random_index], good_activations=best_activations[i])
            
            if mutations is True:

               mutation_prob = random.uniform(0, 1)

               if mutation_prob > bad_genoms_mutation_prob:
                  if (save_best_genom == True and not np.array_equal(best_weights[i], best_weight)) or save_best_genom == False:
                     best_weights[i], best_activations[i] = mutation(best_weights[i], best_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)
                  
               elif mutation_prob < bad_genoms_mutation_prob:
                  bad_weights[i], bad_activations[i] = mutation(bad_weights[i], bad_activations[i], activation_mutate_prob=activation_mutate_prob, activation_add_prob=activation_add_prob, activation_delete_prob=activation_delete_prob, activation_change_prob=activation_change_prob, weight_mutate_prob=weight_mutate_prob, threshold=weight_mutate_rate, dtype=dtype)
      
      if bar_status: progress.update(1)

   weights = np.vstack((bad_weights, best_weights))
   activation_potentiations = bad_activations + best_activations

   ### INFO PRINTING CONSOLE
   
   if show_info == True:
      print("\nGENERATION:", str(what_gen) + ' FINISHED \n')
      print("*** Configuration Settings ***")
      print("  POPULATION SIZE: ", str(len(weights)))
      print("  STRATEGY: ", strategy)

      if strategy == 'cross_over':
         print("  CROSS OVER MODE: ", cross_over_mode)

      print("  POLICY: ", policy)
      print("  MUTATIONS: ", str(mutations))
      print("  BAD GENOMES MUTATION PROB: ", str(bad_genoms_mutation_prob))
      print("  GOOD GENOMES MUTATION PROB: ", str(round(1 - bad_genoms_mutation_prob, 2)))
      print("  WEIGHT MUTATE PROB: ", str(weight_mutate_prob))
      print("  WEIGHT MUTATE RATE (THRESHOLD VALUE FOR SINGLE MUTATION): ", str(weight_mutate_rate))
      print("  ACTIVATION MUTATE PROB: ", str(activation_mutate_prob))
      print("  ACTIVATION ADD PROB: ", str(activation_add_prob))
      print("  ACTIVATION DELETE PROB: ", str(activation_delete_prob))
      print("  ACTIVATION CHANGE PROB: ", str(activation_change_prob))
      print("  ACTIVATION SELECTION ADD PROB: ", str(activation_selection_add_prob))
      print("  ACTIVATION SELECTION CHANGE PROB: ", str(activation_selection_change_prob))
      print("  ACTIVATION SELECTION RATE (THRESHOLD VALUE FOR SINGLE CROSS OVER):", str(activation_selection_rate) + '\n')

      print("*** Performance ***")
      print("  MAX FITNESS: ", str(round(max(fitness), 2)))
      print("  MEAN FITNESS: ", str(round(np.mean(fitness), 2)))
      print("  MIN FITNESS: ", str(round(min(fitness), 2)) + '\n')

      print("  BEST GENOME INDEX: ", str(len(weights)-1))
      print("  NOTE: Genomes are always sorted from the least successful to the most successful according to their performance ranking. Therefore, the genome at the last index is the king of the previous generation. " + '\n')
      

   return weights, activation_potentiations


def evaluate(x_population, weights, activation_potentiations, rl_mode=False, dtype=np.float32):
   """
    Evaluates the performance of a population of genomes, applying different activation functions 
    and weights depending on whether reinforcement learning mode is enabled or not.

    Args:
        x_population (list or numpy.ndarray): A list or 2D numpy array where each element represents
                                               a genome (A list of input features for each genome, or a single set of input features for one genome (only in rl_mode)).
        weights (list or numpy.ndarray): A list or 2D numpy array of weights corresponding to each genome 
                                         in `x_population`. This determines the strength of connections.
        activation_potentiations (list or str): A list where each entry represents an activation function 
                                                or a potentiation strategy applied to each genome. If only one 
                                                activation function is used, this can be a single string.
        rl_mode (bool, optional): If True, reinforcement learning mode is activated, this accepts x_population is a single genom. (Also weights and activation_potentations a single genomes part.)
                                  Default is False.

        dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)                         
         
    Returns:
        list: A list of outputs corresponding to each genome in the population after applying the respective 
              activation function and weights.

    Notes:
        - If `rl_mode` is True:
            - Accepts x_population is a single genom
            - The inputs are flattened, and the activation function is applied across the single genom.
        
        - If `rl_mode` is False:
            - Accepts x_population is a list of genomes
            - Each genome is processed individually, and the results are stored in the `outputs` list.
        
        - `feed_forward()` function is the core function that processes the input with the given weights and activation function.
    
    Example:
        ```python
        outputs = evaluate(x_population, weights, activation_potentiations, rl_mode=False)
        ```

    - The function returns a list of outputs after processing the population, where each element corresponds to 
      the output for each genome in `x_population`.
   """
   
   ### IF RL_MODE IS TRUE, A SINGLE GENOME IS ASSUMED AS INPUT, A FEEDFORWARD PREDICTION IS MADE, AND THE OUTPUT(NPARRAY) IS RETURNED:
   
   ### IF RL_MODE IS FALSE, PREDICTIONS ARE MADE FOR ALL GENOMES IN THE GROUP USING THEIR CORRESPONDING INDEXED INPUTS AND DATA.
   ### THE OUTPUTS ARE RETURNED AS A PYTHON LIST, WHERE EACH GENOME'S OUTPUT MATCHES ITS INDEX:
   
   if rl_mode == True:
      Input = np.array(x_population, copy=False, dtype=dtype)
      Input = Input.ravel()
      
      if isinstance(activation_potentiations, str):
         activation_potentiations = [activation_potentiations]
      
      outputs = feed_forward(Input=Input, is_training=False, activation_potentiation=activation_potentiations, w=weights)
      
   else:
      outputs = [0] * len(x_population)
      for i, genome in enumerate(x_population):

         Input = np.array(genome, copy=False)
         Input = Input.ravel()

         if isinstance(activation_potentiations[i], str):
            activation_potentiations[i] = [activation_potentiations[i]]

         outputs[i] = feed_forward(Input=Input, is_training=False, activation_potentiation=activation_potentiations[i], w=weights[i])

   return outputs


def cross_over(best_weight, good_weight, best_activations, good_activations, cross_over_mode, activation_selection_add_prob, activation_selection_change_prob, activation_selection_rate):
   """
    Performs a selected Crossover operation on two sets of weights and activation functions.
    This function combines two individuals (represented by their weights and activation functions) 
    to create a new individual by exchanging parts of their weight matrices and activation functions.

    Args:
        best_weight (numpy.ndarray): The weight matrix of the first individual (parent).
        good_weight (numpy.ndarray): The weight matrix of the second individual (parent).
        best_activations (str or list): The activation function(s) of the first individual.
        good_activations (str or list): The activation function(s) of the second individual.
        cross_over_mode (str): Determines the crossover method to be used. Options:
                                         - 'tpm': Two-Point Matrix Crossover, where sub-matrices of weights 
                                           are swapped between parents.
                                         - 'plan': Output Connections Crossover, where specific connections 
                                           in the weight matrix are crossed over. Default is 'tpm'.

    Returns:
        tuple: A tuple containing:
            - new_weight (numpy.ndarray): The weight matrix of the new individual created by crossover.
            - new_activations (list): The list of activation functions of the new individual created by crossover.

    Notes:
        - The crossover is performed based on the selected `cross_over_mode`.
            - In 'tpm', random sub-matrices from the parent weight matrices are swapped.
            - In 'plantic', specific connections in the weight matrix are swapped between parents.
        - The crossover operation combines the activation functions of both parents:
            - If the activation functions are passed as strings, they are converted to lists for uniform handling.
            - The resulting activation functions depend on the crossover method and the parent's configuration.

    Example:
        ```python
        new_weights, new_activations = cross_over(best_weight, good_weight, best_activations, good_activations, cross_over_mode='tpm')
        ```
    """
   
   ### THE GIVEN GENOMES' WEIGHTS ARE RANDOMLY SELECTED AND COMBINED OVER A RANDOM RANGE. SIMILARLY, THEIR ACTIVATIONS ARE COMBINED. A NEW GENOME IS RETURNED WITH THE COMBINED WEIGHTS FIRST, FOLLOWED BY THE ACTIVATIONS:
   
   start = 0
   
   row_end = best_weight.shape[0]
   col_end = best_weight.shape[1]

   while True:

      row_cut_start = int(random.uniform(start, row_end))
      col_cut_start = int(random.uniform(start, col_end))

      row_cut_end = int(random.uniform(start, row_end))
      col_cut_end = int(random.uniform(start, col_end))

      if (row_cut_end > row_cut_start) and (col_cut_end > col_cut_start):
         break

   new_weight = np.copy(best_weight)
   best_w2 = np.copy(good_weight)
   
   if cross_over_mode == 'tpm':
      new_weight[row_cut_start:row_cut_end, col_cut_start:col_cut_end] = best_w2[row_cut_start:row_cut_end, col_cut_start:col_cut_end]

   elif cross_over_mode == 'plantic':
      new_weight[row_cut_start:row_cut_end,:] = best_w2[row_cut_start:row_cut_end,:]


   if isinstance(best_activations, str):
      best = [best_activations]

   if isinstance(good_activations, str):
      good = [good_activations]

   if isinstance(best_activations, list):
      best = best_activations

   if isinstance(good_activations, list):
      good = good_activations

   new_activations = list(np.copy(best))

   activation_selection_add_prob = 1 - activation_selection_add_prob # if prob 0.8 (%80) then 1 - 0.8. Because 0-1 random number probably greater than 0.2
   potential_activation_selection_add = random.uniform(0, 1)

   if potential_activation_selection_add > activation_selection_add_prob:
      
      new_threshold = activation_selection_rate
      
      while True:
         
         random_index_good = int(random.uniform(0, len(good)-1))
         random_good_activation = good[random_index_good]

         new_activations.append(random_good_activation)

         if len(best) > new_threshold:
            new_threshold += activation_selection_rate
            pass
         
         else:
            break

   activation_selection_change_prob = 1 - activation_selection_change_prob
   potential_activation_selection_change_prob = random.uniform(0, 1)

   if potential_activation_selection_change_prob > activation_selection_change_prob:

      new_threshold = activation_selection_rate
      
      while True:
         
         random_index_good = int(random.uniform(0, len(good)-1))
         random_index_best = int(random.uniform(0, len(best)-1))
         random_good_activation = good[random_index_good]

         new_activations[random_index_best] = good[random_index_good]

         if len(best) > new_threshold:
            new_threshold += activation_selection_rate
            pass
         
         else:
            break

   return new_weight, new_activations

def potentiate(best_weight, good_weight, best_activations, good_activations, dtype=np.float32):
   """
    Combines two sets of weights and activation functions by adding the weight matrices and 
    concatenating the activation functions. The resulting weight matrix is normalized. (Max abs normalization.)
    
    Args:
        best_weight (numpy.ndarray): The weight matrix of the first individual (parent).
        good_weight (numpy.ndarray): The weight matrix of the second individual (parent).
        best_activations (str or list): The activation function(s) of the first individual.
        good_activations (str or list): The activation function(s) of the second individual.
        dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)
    
    Returns:
        tuple: A tuple containing:
            - new_weight (numpy.ndarray): The new weight matrix after potentiation and normalization. (Max abs normalization.)
            - new_activations (list): The new activation functions after concatenation.
    
    Notes:
        - The weight matrices are element-wise added and then normalized using the `normalization` function. (Max abs normalization.)
        - The activation functions from both parents are concatenated to form the new activation functions list.
        - If the activation functions are passed as strings, they are converted to lists for uniform handling.
   """
    
   new_weight = best_weight + good_weight
   new_weight = normalization(new_weight, dtype=dtype)
   
   if isinstance(best_activations, str):
      best = [best_activations]

   if isinstance(good_activations, str):
      good = [good_activations]

   if isinstance(best_activations, list):
      best = best_activations

   if isinstance(good_activations, list):
      good = good_activations
   
   new_activations = best + good
   
   return new_weight, new_activations

def mutation(weight, activations, activation_mutate_prob, activation_add_prob, activation_delete_prob, activation_change_prob, weight_mutate_prob, threshold, dtype=np.float32):
   """
    Performs mutation on the given weight matrix and activation functions.
    - The weight matrix is mutated by randomly changing its values based on the mutation probability.
    - The activation functions are mutated by adding, removing, or replacing them with predefined probabilities.

    Args:
        weight (numpy.ndarray): The weight matrix to mutate.
        activations (list): The list of activation functions to mutate.
        activation_mutate_prob (float): The overall probability of mutating activation functions.
        activation_add_prob (float): Probability of adding a new activation function.
        activation_delete_prob (float): Probability of removing an existing activation function.
        activation_change_prob (float): Probability of replacing an existing activation function with a new one.
        weight_mutate_prob (float): The probability of mutating weight matrix.
        threshold (float): If the value you enter here is equal to the result of input layer * output layer, only a single weight will be mutated during each mutation process. If the value you enter here is half of the result of input layer * output layer, two weights in the weight matrix will be mutated.
         dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

    Returns:
        tuple: A tuple containing:
            - mutated_weight (numpy.ndarray): The weight matrix after mutation.
            - mutated_activations (list): The list of activation functions after mutation.

    Notes:
        - Weight mutation:
            - Each weight has a chance defined by `weight_mutate_prob` to be altered by adding a random value 
              within the range of [0, 1].
        - Activation mutation:
            - If `activation_mutate_prob` is triggered, one or more activation functions can be added, removed, 
              or replaced based on their respective probabilities (`activation_add_prob`, `activation_delete_prob`, 
              `activation_change_prob`).
        - The mutation probabilities should be chosen carefully to balance exploration and exploitation during 
          the optimization process.
   """

   if isinstance(activations, str):
      activations = [activations]

   weight_mutate_prob = 1 - weight_mutate_prob # if prob 0.8 (%80) then 1 - 0.8. Because 0-1 random number probably greater than 0.2
   potential_weight_mutation = random.uniform(0, 1)

   if potential_weight_mutation > weight_mutate_prob:

      start = 0
      row_end = weight.shape[0]
      col_end = weight.shape[1]
      new_threshold = threshold
      
      while True:
         
         selected_row = int(random.uniform(start, row_end))
         selected_col = int(random.uniform(start, col_end))

         weight[selected_row, selected_col] = random.uniform(-1, 1)

         if int(row_end * col_end) > new_threshold:
            new_threshold += threshold
            pass
         
         else:
            break 
   
   
   activation_mutate_prob = 1 - activation_mutate_prob
   potential_activation_mutation = random.uniform(0, 1)

   if potential_activation_mutation > activation_mutate_prob:
      
      except_this = ['spiral', 'circular']
      all_acts = [item for item in all_activations() if item not in except_this] # SPIRAL AND CIRCULAR ACTIVATION DISCARDED
      
      activation_add_prob = 1 - activation_add_prob
      activation_delete_prob = 1 - activation_delete_prob
      activation_change_prob = 1 - activation_change_prob

      potential_activation_add_prob = random.uniform(0, 1)
      potential_activation_delete_prob = random.uniform(0, 1)
      potential_activation_change_prob = random.uniform(0, 1)

      if potential_activation_add_prob > activation_add_prob:

         try:
            
            random_index_all_act = int(random.uniform(0, len(all_acts)-1))
            activations.append(all_acts[random_index_all_act])
            
            for i in range(weight.shape[0]):

               weight[i,:] = apply_activation(weight[i,:], activations[-1])

            weight = normalization(weight, dtype=dtype)
            
         except:

            activation = activations
            activations = []

            activations.append(activation)
            activations.append(all_acts[int(random.uniform(0, len(all_acts)-1))])
            
            for i in range(weight.shape[0]):

               weight[i,:] = apply_activation(weight[i,:], activations[-1])

            weight = normalization(weight, dtype=dtype)
            
      if potential_activation_delete_prob > activation_delete_prob and len(activations) > 1:
         
         random_index = random.randint(0, len(activations) - 1)
         
         wc = np.copy(weight)
         for i in range(weight.shape[0]):

            wc[i,:] = apply_activation(wc[i,:], activations[random_index])
            weight[i,:] -= wc[i,:]

         activations.pop(random_index)
         weight = normalization(weight, dtype=dtype)
         
            
      if potential_activation_change_prob > activation_change_prob:
         
         random_index_all_act = int(random.uniform(0, len(all_acts)-1))
         random_index_genom_act = int(random.uniform(0, len(activations)-1))

         activations[random_index_genom_act] = all_acts[random_index_all_act]

         wc = np.copy(weight)
         for i in range(weight.shape[0]):

            wc[i,:] = apply_activation(wc[i,:], activations[random_index_genom_act])
            weight[i,:] -= wc[i,:]

         weight = normalization(weight, dtype=dtype)

         for i in range(weight.shape[0]):

            weight[i,:] = apply_activation(weight[i,:], activations[random_index_genom_act])

         weight = normalization(weight, dtype=dtype)

   return weight, activations


import numpy as np
from colorama import Fore, Style
import sys
from datetime import datetime
import pickle
from scipy import io
import scipy.io as sio
import pandas as pd


def save_model(model_name,
               W,
               scaler_params=None,
               model_type='PLAN',
               test_acc=None,
               model_path='',
               activation_potentiation=['linear'],
               weights_type='npy',
               weights_format='raw',
               show_architecture=False,
               show_info=True
               ):

    """
    Function to save a potentiation learning artificial neural network model.

    Arguments:

    model_name (str): Name of the model.

    model_type (str): Type of the model. default: 'PLAN'

    test_acc (float): Test accuracy of the model. default: None

    weights_type (str): Type of weights to save (options: 'txt', 'pkl', 'npy', 'mat'). default: 'npy'

    WeightFormat (str): Format of the weights (options: 'f', 'raw'). default: 'raw'

    model_path (str): Path where the model will be saved. For example: C:/Users/beydili/Desktop/denemePLAN/ default: ''

    scaler_params (list[num, num]): standard scaler params list: mean,std. If not used standard scaler then be: None.

    W: Weights of the model.

    activation_potentiation (list): For deeper PLAN networks, activation function parameters. For more information please run this code: plan.activations_list() default: ['linear']

    show_architecture (bool): It draws model architecture. True or False. Default: False

    show_info (bool): Prints model details into console. default: True

    Returns:
    No return.
    """
    
    from .visualizations import draw_model_architecture

    class_count = W.shape[0]

    if test_acc != None:
        test_acc= float(test_acc)

    if weights_type != 'txt' and weights_type != 'npy' and weights_type != 'mat' and weights_type != 'pkl':
        print(Fore.RED + "ERROR110: Save Weight type (File Extension) Type must be 'txt' or 'npy' or 'mat' or 'pkl' from: save_model" + Style.RESET_ALL)
        sys.exit()

    if weights_format != 'd' and weights_format != 'f' and weights_format != 'raw':
        print(Fore.RED + "ERROR111: Weight Format Type must be 'd' or 'f' or 'raw' from: save_model" + Style.RESET_ALL)
        sys.exit()

    NeuronCount = 0
    SynapseCount = 0


    try:
            NeuronCount += np.shape(W)[0] + np.shape(W)[1]
            SynapseCount += np.shape(W)[0] * np.shape(W)[1]
    except:

        print(Fore.RED + "ERROR: Weight matrices has a problem from: save_model" + Style.RESET_ALL)
        sys.exit()

    if scaler_params != None:

        if len(scaler_params) > len(activation_potentiation):

            activation_potentiation += ['']

        elif len(activation_potentiation) > len(scaler_params):

            for i in range(len(activation_potentiation) - len(scaler_params)):

                scaler_params.append(' ')

    data = {'MODEL NAME': model_name,
            'MODEL TYPE': model_type,
            'CLASS COUNT': class_count,
            'NEURON COUNT': NeuronCount,
            'SYNAPSE COUNT': SynapseCount,
            'TEST ACCURACY': test_acc,
            'SAVE DATE': datetime.now(),
            'WEIGHTS TYPE': weights_type,
            'WEIGHTS FORMAT': weights_format,
            'MODEL PATH': model_path,
            'STANDARD SCALER': scaler_params,
            'ACTIVATION POTENTIATION': activation_potentiation
            }

    df = pd.DataFrame(data)
    df.to_pickle(model_path + model_name + '.pkl')


    try:

        if weights_type == 'txt' and weights_format == 'f':

                np.savetxt(model_path + model_name + '_weights.txt',  W, fmt='%f')

        if weights_type == 'txt' and weights_format == 'raw':

                np.savetxt(model_path + model_name + '_weights.txt',  W)

        ###

        
        if weights_type == 'pkl' and weights_format == 'f':

            with open(model_path + model_name + '_weights.pkl', 'wb') as f:
                pickle.dump(W.astype(float), f)

        if weights_type == 'pkl' and weights_format =='raw':
        
            with open(model_path + model_name + '_weights.pkl', 'wb') as f:
                pickle.dump(W, f)

        ###

        if weights_type == 'npy' and weights_format == 'f':

                np.save(model_path + model_name + '_weights.npy',  W, W.astype(float))

        if weights_type == 'npy' and weights_format == 'raw':

                np.save(model_path + model_name + '_weights.npy',  W)

        ###

        if weights_type == 'mat' and weights_format == 'f':

                w = {'w': W.astype(float)}
                io.savemat(model_path + model_name + '_weights.mat', w)

        if weights_type == 'mat' and weights_format == 'raw':
                
                w = {'w': W}
                io.savemat(model_path + model_name + '_weights.mat', w)

    except:

        print(Fore.RED + "ERROR: Model Weights not saved. Check the Weight parameters. SaveFilePath expl: 'C:/Users/hasancanbeydili/Desktop/denemePLAN/' from: save_model" + Style.RESET_ALL)
        sys.exit()
    
    if show_info:
        print(df)
    
        message = (
            Fore.GREEN + "Model Saved Successfully\n" +
            Fore.MAGENTA + "Don't forget, if you want to load model: model log file and weight files must be in the same directory." +
            Style.RESET_ALL
        )
        
        print(message)

    if show_architecture:
        draw_model_architecture(model_name=model_name, model_path=model_path)



def load_model(model_name,
               model_path,
               ):
    """
   Function to load a potentiation learning model.

   Arguments:

   model_name (str): Name of the model.

   model_path (str): Path where the model is saved.

   Returns:
   lists: W(list[num]), activation_potentiation, DataFrame of the model
    """

    try:

         df = pd.read_pickle(model_path + model_name + '.pkl')

    except:

        print(Fore.RED + "ERROR: Model Path error. acceptable form: 'C:/Users/hasancanbeydili/Desktop/denemePLAN/' from: load_model" + Style.RESET_ALL)

        sys.exit()

    activation_potentiation = list(df['ACTIVATION POTENTIATION'])
    activation_potentiation = [x for x in activation_potentiation if not (isinstance(x, float) and np.isnan(x))]
    activation_potentiation = [item for item in activation_potentiation if item != ''] 

    scaler_params = df['STANDARD SCALER'].tolist()
    
    try:
        if scaler_params[0] == None:
            scaler_params = scaler_params[0]

    except:
        scaler_params = [item for item in scaler_params if isinstance(item, np.ndarray)]

     
    model_name = str(df['MODEL NAME'].iloc[0])
    WeightType = str(df['WEIGHTS TYPE'].iloc[0])

    if WeightType == 'txt':
            W = np.loadtxt(model_path + model_name + '_weights.txt')
    elif WeightType == 'npy':
            W = np.load(model_path + model_name + '_weights.npy')
    elif WeightType == 'mat':
            W = sio.loadmat(model_path + model_name + '_weights.mat')
    elif WeightType == 'pkl':
        with open(model_path + model_name + '_weights.pkl', 'rb') as f:
            W = pickle.load(f)
    else:

        raise ValueError(
            Fore.RED + "Incorrect weight type value. Value must be 'txt', 'npy', 'pkl' or 'mat' from: load_model." + Style.RESET_ALL)
        
    if WeightType == 'mat':
        W = W['w']

    return W, None, None, activation_potentiation, scaler_params


def predict_model_ssd(Input, model_name, model_path='', dtype=np.float32):

    """
    Function to make a prediction using a potentiation learning artificial neural network (PLAN).
    from storage

    Arguments:

    Input (list or ndarray): Input data for the model (single vector or single matrix).

    model_name (str): Name of the model.

    model_path (str): Path of the model. Default: ''

    dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!]

    Returns:
    ndarray: Output from the model.
    """
    
    from .plan import feed_forward
    from .data_operations import standard_scaler
    
    model = load_model(model_name, model_path)
    
    activation_potentiation = model[get_act_pot()]
    scaler_params = model[get_scaler()]
    W = model[get_weights()]

    Input = standard_scaler(None, Input, scaler_params, dtype=dtype)

    neural_layer = Input
    neural_layer = np.array(neural_layer, copy=False)
    neural_layer = neural_layer.ravel()

    try:
        neural_layer = feed_forward(neural_layer, np.copy(W.astype(dtype, copy=False)), is_training=False, Class='?', activation_potentiation=activation_potentiation)
        return neural_layer
    except:
        print(Fore.RED + "ERROR: Unexpected Output or wrong model parameters from: predict_model_ssd." + Style.RESET_ALL)
        sys.exit()
    

def reverse_predict_model_ssd(output, model_name, model_path='', dtype=np.float32):

    """
    reverse prediction function from storage
    Arguments:

    output (list or ndarray): output layer for the model (single probability vector, output layer of trained model).

    model_name (str): Name of the model.

    model_path (str): Path of the model. Default: ''

    dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!]
    
    Returns:
    ndarray: Input from the model.
    """
    
    model = load_model(model_name, model_path)
    
    W = model[get_weights()]

    try:
        Input = np.dot(output.astype(dtype, copy=False), np.copy(W.astype(dtype, copy=False)))
        return Input
    except:
        print(Fore.RED + "ERROR: Unexpected Output or wrong model parameters from: reverse_predict_model_ssd." + Style.RESET_ALL)
        sys.exit()
    


def predict_model_ram(Input, W, scaler_params=None, activation_potentiation=['linear'], dtype=np.float32):

    """
    Function to make a prediction using a potentiation learning artificial neural network (PLAN).
    from memory.

    Arguments:

    Input (list or ndarray): Input data for the model (single vector or single matrix).

    W (list of ndarrays): Weights of the model.

    scaler_params (list): standard scaler params list: mean,std. (optional) Default: None.
    
    activation_potentiation (list): ac list for deep PLAN. default: [None] ('linear') (optional)

    dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!]

    Returns:
    ndarray: Output from the model.
    """

    from .data_operations import standard_scaler
    from .plan import feed_forward

    Input = standard_scaler(None, Input, scaler_params, dtype=dtype)
    
    try:
        
        neural_layer = Input
        neural_layer = np.array(neural_layer, copy=False)
        neural_layer = neural_layer.ravel()

        neural_layer = feed_forward(neural_layer, np.copy(W.astype(dtype, copy=False)), is_training=False, Class='?', activation_potentiation=activation_potentiation)

        return neural_layer
    
    except:
        print(Fore.RED + "ERROR: Unexpected input or wrong model parameters from: predict_model_ram." + Style.RESET_ALL)
        sys.exit()

def reverse_predict_model_ram(output, W, dtype=np.float32):

    """
    reverse prediction function from memory

    Arguments:

    output (list or ndarray): output layer for the model (single probability vector, output layer of trained model).

    W (list of ndarrays): Weights of the model.

    dtype (numpy.dtype): Data type for the arrays. np.float32 by default. Example: np.float64 or np.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!]

    Returns:
    ndarray: Input from the model.
    """

    try:
        Input = np.dot(output.astype(dtype, copy=False), np.copy(W.astype(dtype, copy=False)))
        return Input
    
    except:
        print(Fore.RED + "ERROR: Unexpected Output or wrong model parameters from: reverse_predict_model_ram." + Style.RESET_ALL)
        sys.exit()


def get_weights():

    return 0


def get_preds():

    return 1


def get_acc():

    return 2


def get_act_pot():

    return 3


def get_scaler():

    return 4

def get_preds_softmax():

    return 5


from tqdm import tqdm

def loading_bars():
    
    GREEN = "\033[92m"
    RESET = "\033[0m"

    bar_format_normal = f"{GREEN}{{bar}}{GREEN} {RESET} {{l_bar}} {{remaining}} {{postfix}}"
    bar_format_learner = f"{GREEN}{{bar}}{GREEN} {RESET} {{remaining}} {{postfix}}"

    return bar_format_normal, bar_format_learner


def initialize_loading_bar(total, desc, ncols, bar_format, leave=True):
    return tqdm(
        total=total,
        leave=leave,
        desc=desc,
        ascii="▱▰",
        bar_format=bar_format,
        ncols=ncols,
    )



import subprocess
subprocess.check_call(["pip", "install", 'setuptools==75.6.0'])
import pkg_resources

print("Auto checking and installation dependencies for PyerualJetwork")

package_names = [
    'scipy==1.13.1',
    'tqdm==4.66.4',
    'seaborn==0.13.2',
    'pandas==2.2.2',
    'networkx==3.3',
    'numpy==1.26.4',
    'matplotlib==3.9.0',
    'colorama==0.4.6',
    'psutil==6.1.1',
    'cupy-cuda12x==13.3.0'
]

installed_packages = pkg_resources.working_set
installed = {pkg.key: pkg.version for pkg in installed_packages}
err = 0

for package_name in package_names:
    package_name_only, required_version = package_name.split('==')
    
    if package_name_only not in installed:

        try:
            print(f"{package_name} Installing...")
            subprocess.check_call(["pip", "install", package_name])
        except Exception as e:
            err += 1
            print(f"Error installing {package_name} library, installation continues: {e}")
    else:

        installed_version = installed[package_name_only]
        if installed_version != required_version:
            print(f"Updating {package_name_only} from version {installed_version} to {required_version}...")
            try:
                subprocess.check_call(["pip", "install", package_name])
            except Exception as e:
                err += 1
                print(f"Error updating {package_name} library, installation continues: {e}")
        else:
            print(f"{package_name} ready.")

print(f"PyerualJetwork is ready to use with {err} errors")

__version__ = "4.1.8b8"
__update__ = "* Changes: https://github.com/HCB06/PyerualJetwork/blob/main/CHANGES\n* PyerualJetwork document: https://github.com/HCB06/PyerualJetwork/blob/main/Welcome_to_PyerualJetwork/PYERUALJETWORK_USER_MANUEL_AND_LEGAL_INFORMATION(EN).pdf\n* YouTube tutorials: https://www.youtube.com/@HasanCanBeydili"

def print_version(__version__):
    print(f"PyerualJetwork Version {__version__}" + '\n')

def print_update_notes(__update__):
    print(f"Update Notes:\n{__update__}")

print_version(__version__)
print_update_notes(__update__)


import cupy as cp
from scipy.special import expit, softmax
import warnings

# ACTIVATION FUNCTIONS ----

def all_activations():
    
    activations_list = ['linear', 'sigmoid', 'relu', 'tanh', 'circular', 'spiral', 'swish', 'sin_plus', 'mod_circular', 'tanh_circular', 'leaky_relu', 'softplus', 'elu', 'gelu', 'selu', 'sinakt', 'p_squared', 'sglu', 'dlrelu', 'exsig', 'acos', 'gla', 'srelu', 'qelu', 'isra', 'waveakt', 'arctan', 'bent_identity', 'sech', 'softsign', 'pwl', 'cubic', 'gaussian', 'sine', 'tanh_square', 'mod_sigmoid', 'quartic', 'square_quartic', 'cubic_quadratic', 'exp_cubic', 'sine_square', 'logarithmic', 'scaled_cubic', 'sine_offset']
    
    return activations_list

def spiral_activation(x):

    r = cp.sqrt(cp.sum(x**2))
    
    theta = cp.arctan2(x[1:], x[:-1])

    spiral_x = r * cp.cos(theta + r)
    spiral_y = r * cp.sin(theta + r)


    spiral_output = cp.concatenate([cp.array([spiral_x[0]]), spiral_y])
    
    return spiral_output


def Softmax(
    x  # num: Input data to be transformed using softmax function.
):
    """
    Applies the softmax function to the input data.

    Args:
        (num): Input data to be transformed using softmax function.

    Returns:
       (num): Transformed data after applying softmax function.
    """
    
    return cp.array(softmax(x.get()))


def Sigmoid(
    x  # num: Input data to be transformed using sigmoid function.
):
    """
    Applies the sigmoid function to the input data.

    Args:
        (num): Input data to be transformed using sigmoid function.

    Returns:
        (num): Transformed data after applying sigmoid function.
    """
    return expit(x)


def Relu(
    x  # num: Input data to be transformed using ReLU function.
):
    """
    Applies the Rectified Linear Unit (ReLU) function to the input data.

    Args:
        (num): Input data to be transformed using ReLU function.

    Returns:
        (num): Transformed data after applying ReLU function.
    """

    return cp.maximum(0, x)


def tanh(x):
    return cp.tanh(x)

def swish(x):
    return x * (1 / (1 + cp.exp(-x)))

def sin_plus(x):
    return (cp.sin(x) + 1) / 2

def modular_circular_activation(x, period=2*cp.pi):
    return cp.mod(x, period) / period

def tanh_circular_activation(x):
    return (cp.tanh(x) + 1) / 2

def leaky_relu(x, alpha=0.01):
    return cp.where(x > 0, x, alpha * x)

def softplus(x):
    return cp.log(1 + cp.exp(x))

def elu(x, alpha=1.0):
    return cp.where(x > 0, x, alpha * (cp.exp(x) - 1))

def gelu(x):
    return 0.5 * x * (1 + cp.tanh(cp.sqrt(2 / cp.pi) * (x + 0.044715 * cp.power(x, 3))))

def selu(x, lambda_=1.0507, alpha=1.6733):
    return lambda_ * cp.where(x > 0, x, alpha * (cp.exp(x) - 1))

def sinakt(x):
    return cp.sin(x) + cp.cos(x)

def p_squared(x, alpha=1.0, beta=0.0):
    return alpha * x**2 + beta * x

def sglu(x, alpha=1.0):
    return cp.array(softmax(alpha * x.get())) * x

# 4. Double Leaky ReLU (DLReLU)
def dlrelu(x):
    return cp.maximum(0.01 * x, x) + cp.minimum(0.01 * x, 0.1 * x)

# 5. Exponential Sigmoid (ExSig)
def exsig(x):
    return 1 / (1 + cp.exp(-x**2))

# 6. Adaptive Cosine Activation (ACos)
def acos(x, alpha=1.0, beta=0.0):
    return cp.cos(alpha * x + beta)

# 7. Gaussian-like Activation (GLA)
def gla(x, alpha=1.0, mu=0.0):
    return cp.exp(-alpha * (x - mu)**2)

# 8. Swish ReLU (SReLU)
def srelu(x):
    return x * (1 / (1 + cp.exp(-x))) + cp.maximum(0, x)

# 9. Quadratic Exponential Linear Unit (QELU)
def qelu(x):
    return x**2 * cp.exp(x) - 1

# 10. Inverse Square Root Activation (ISRA)
def isra(x):
    return x / cp.sqrt(cp.abs(x) + 1)

def waveakt(x, alpha=1.0, beta=2.0, gamma=3.0):
    return cp.sin(alpha * x) * cp.cos(beta * x) * cp.sin(gamma * x)

def arctan(x):
    return cp.arctan(x)

def bent_identity(x):
    return (cp.sqrt(x**2 + 1) - 1) / 2 + x

def circular_activation(x, scale=2.0, frequency=1.0, shift=0.0):    
    
    n_features = x.shape[0]
    
    circular_output = cp.zeros_like(x)
    
    for i in range(n_features):
        
        r = cp.sqrt(cp.sum(x**2))
        theta = 2 * cp.pi * (i / n_features) + shift
        
        circular_x = r * cp.cos(theta + frequency * r) * scale
        circular_y = r * cp.sin(theta + frequency * r) * scale
        
        if i % 2 == 0:
            circular_output[i] = circular_x
        else:
            circular_output[i] = circular_y
    
    return circular_output

def sech(x):
    return 2 / (cp.exp(x) + cp.exp(-x))

def softsign(x):
    return x / (1 + cp.abs(x))

def pwl(x, alpha=0.5, beta=1.5):
    return cp.where(x <= 0, alpha * x, beta * x)

def cubic(x):
    return x**3

def gaussian(x, alpha=1.0, mu=0.0):
    return cp.exp(-alpha * (x - mu)**2)
                
def sine(x, alpha=1.0):
    return cp.sin(alpha * x)

def tanh_square(x):
    return cp.tanh(x)**2

def mod_sigmoid(x, alpha=1.0, beta=0.0):
    return 1 / (1 + cp.exp(-alpha * x + beta))

def quartic(x):
    return x**4

def square_quartic(x):
    return (x**2)**2

def cubic_quadratic(x):
    return x**3 * (x**2)

def exp_cubic(x):
    return cp.exp(x**3)

def sine_square(x):
    return cp.sin(x)**2

def logarithmic(x):
    return cp.log(x**2 + 1)

def scaled_cubic(x, alpha=1.0):
    return alpha * x**3

def sine_offset(x, beta=0.0):
    return cp.sin(x + beta)



def safe_add(current_sum, new_value):
        try:
            return current_sum + new_value
        except OverflowError:
            return cp.array(current_sum) + cp.array(new_value)
    
def apply_activation(Input, activation_list):
    """
        Applies a sequence of activation functions to the input.
        
        Args:
            Input (numpy.ndarray): The input to apply activations to.
            activation_list (list): A list of activation function names to apply.
        
        Returns:
            numpy.ndarray: The input after all activations have been applied.
    """
    
    origin_input = cp.copy(Input)

    for i in range(len(activation_list)):
        try:
            if activation_list[i] == 'sigmoid':
                Input = safe_add(Input, Sigmoid(origin_input))
            elif activation_list[i] == 'swish':
                Input = safe_add(Input, swish(origin_input))
            elif activation_list[i] == 'mod_circular':
                Input = safe_add(Input, modular_circular_activation(origin_input))
            elif activation_list[i] == 'tanh_circular':
                Input = safe_add(Input, tanh_circular_activation(origin_input))
            elif activation_list[i] == 'leaky_relu':
                Input = safe_add(Input, leaky_relu(origin_input))
            elif activation_list[i] == 'relu':
                Input = safe_add(Input, Relu(origin_input))
            elif activation_list[i] == 'softplus':
                Input = safe_add(Input, softplus(origin_input))
            elif activation_list[i] == 'elu':
                Input = safe_add(Input, elu(origin_input))
            elif activation_list[i] == 'gelu':
                Input = safe_add(Input, gelu(origin_input))
            elif activation_list[i] == 'selu':
                Input = safe_add(Input, selu(origin_input))
            elif activation_list[i] == 'tanh':
                Input = safe_add(Input, tanh(origin_input))
            elif activation_list[i] == 'sinakt':
                Input = safe_add(Input, sinakt(origin_input))
            elif activation_list[i] == 'p_squared':
                Input = safe_add(Input, p_squared(origin_input))
            elif activation_list[i] == 'sglu':
                Input = safe_add(Input, sglu(origin_input, alpha=1.0))
            elif activation_list[i] == 'dlrelu':
                Input = safe_add(Input, dlrelu(origin_input))
            elif activation_list[i] == 'exsig':
                Input = safe_add(Input, exsig(origin_input))
            elif activation_list[i] == 'sin_plus':
                Input = safe_add(Input, sin_plus(origin_input))
            elif activation_list[i] == 'acos':
                Input = safe_add(Input, acos(origin_input, alpha=1.0, beta=0.0))
            elif activation_list[i] == 'gla':
                Input = safe_add(Input, gla(origin_input, alpha=1.0, mu=0.0))
            elif activation_list[i] == 'srelu':
                Input = safe_add(Input, srelu(origin_input))
            elif activation_list[i] == 'qelu':
                Input = safe_add(Input, qelu(origin_input))
            elif activation_list[i] == 'isra':
                Input = safe_add(Input, isra(origin_input))
            elif activation_list[i] == 'waveakt':
                Input = safe_add(Input, waveakt(origin_input))
            elif activation_list[i] == 'arctan':
                Input = safe_add(Input, arctan(origin_input))
            elif activation_list[i] == 'bent_identity':
                Input = safe_add(Input, bent_identity(origin_input))
            elif activation_list[i] == 'sech':
                Input = safe_add(Input, sech(origin_input))
            elif activation_list[i] == 'softsign':
                Input = safe_add(Input, softsign(origin_input))
            elif activation_list[i] == 'pwl':
                Input = safe_add(Input, pwl(origin_input))
            elif activation_list[i] == 'cubic':
                Input = safe_add(Input, cubic(origin_input))
            elif activation_list[i] == 'gaussian':
                Input = safe_add(Input, gaussian(origin_input))
            elif activation_list[i] == 'sine':
                Input = safe_add(Input, sine(origin_input))
            elif activation_list[i] == 'tanh_square':
                Input = safe_add(Input, tanh_square(origin_input))
            elif activation_list[i] == 'mod_sigmoid':
                Input = safe_add(Input, mod_sigmoid(origin_input))
            elif activation_list[i] == 'linear':
                Input = safe_add(Input, origin_input)
            elif activation_list[i] == 'quartic':
                Input = safe_add(Input, quartic(origin_input))
            elif activation_list[i] == 'square_quartic':
                Input = safe_add(Input, square_quartic(origin_input))
            elif activation_list[i] == 'cubic_quadratic':
                Input = safe_add(Input, cubic_quadratic(origin_input))
            elif activation_list[i] == 'exp_cubic':
                Input = safe_add(Input, exp_cubic(origin_input))
            elif activation_list[i] == 'sine_square':
                Input = safe_add(Input, sine_square(origin_input))
            elif activation_list[i] == 'logarithmic':
                Input = safe_add(Input, logarithmic(origin_input))
            elif activation_list[i] == 'scaled_cubic':
                Input = safe_add(Input, scaled_cubic(origin_input, 1.0))
            elif activation_list[i] == 'sine_offset':
                Input = safe_add(Input, sine_offset(origin_input, 1.0))
            elif activation_list[i] == 'spiral':
                Input = safe_add(Input, spiral_activation(origin_input))
            elif activation_list[i] == 'circular':
                Input = safe_add(Input, circular_activation(origin_input))
        
        except Exception as e:
            warnings.warn(f"Error in activation {activation_list[i]}: {str(e)}", RuntimeWarning)
            if not isinstance(Input, cp.ndarray):
                Input = cp.array(Input)
            if not isinstance(origin_input, cp.ndarray):
                origin_input = cp.array(origin_input)
            continue
 
    return Input


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Setting Up
setup(
    name="pyerualjetwork",
    version="4.1.8b8",
    author="Hasan Can Beydili",
    author_email="tchasancan@gmail.com",
    description=(
        "PyerualJetwork is a machine learning library written in Python for professionals, incorporating advanced, unique, new, and modern techniques.\n"
        "* Changes: https://github.com/HCB06/PyerualJetwork/blob/main/CHANGES\n"
        "* PyerualJetwork document: "
        "https://github.com/HCB06/PyerualJetwork/blob/main/Welcome_to_PyerualJetwork/"
        "PYERUALJETWORK_USER_MANUEL_AND_LEGAL_INFORMATION(EN).pdf."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    keywords=[
        "model evaluation",
        "classification",
        "potentiation learning artificial neural networks",
        "NEAT",
        "genetic algorithms",
        "reinforcement learning",
        "neural networks",
    ],
)


# -*- coding: utf-8 -*-
"""

MAIN MODULE FOR PLAN_CUDA

PLAN document: https://github.com/HCB06/PyerualJetwork/blob/main/Welcome_to_PLAN/PLAN.pdf
PYERUALJETWORK document: https://github.com/HCB06/PyerualJetwork/blob/main/Welcome_to_PyerualJetwork/PYERUALJETWORK_USER_MANUEL_AND_LEGAL_INFORMATION(EN).pdf

@author: Hasan Can Beydili
@YouTube: https://www.youtube.com/@HasanCanBeydili
@Linkedin: https://www.linkedin.com/in/hasan-can-beydili-77a1b9270/
@Instagram: https://www.instagram.com/canbeydilj/
@contact: tchasancan@gmail.com
"""

import cupy as cp
import numpy as np
from colorama import Fore
import math

### LIBRARY IMPORTS ###
from .ui import loading_bars, initialize_loading_bar
from .data_operations_cuda import normalization, decode_one_hot, batcher
from .loss_functions_cuda import binary_crossentropy, categorical_crossentropy
from .activation_functions_cuda import apply_activation, Softmax, all_activations
from .metrics_cuda import metrics
from .model_operations_cuda import get_acc, get_preds, get_preds_softmax
from .memory_operations import transfer_to_gpu, transfer_to_cpu, optimize_labels
from .visualizations_cuda import (
    draw_neural_web,
    update_neural_web_for_fit,
    plot_evaluate,
    update_neuron_history,
    initialize_visualization_for_fit,
    update_weight_visualization_for_fit,
    update_decision_boundary_for_fit,
    update_validation_history_for_fit,
    display_visualization_for_fit,
    display_visualizations_for_learner,
    update_history_plots_for_learner,
    initialize_visualization_for_learner,
    update_neuron_history_for_learner,
    show
)

### GLOBAL VARIABLES ###
bar_format_normal = loading_bars()[0]
bar_format_learner = loading_bars()[1]

# BUILD -----

def fit(
    x_train,
    y_train,
    val=False,
    val_count=None,
    activation_potentiation=['linear'],
    x_val=None,
    y_val=None,
    show_training=None,
    interval=100,
    LTD=0,
    decision_boundary_status=True,
    train_bar=True,
    auto_normalization=True,
    neurons_history=False,
    dtype=cp.float32,
    memory='gpu'
):
    """
    Creates a model to fitting data.
    
    fit Args:

        x_train (list[num]): List or numarray of input data.

        y_train (list[num]): List or numarray of target labels. (one hot encoded)

        val (None or True): validation in training process ? None or True default: None (optional)

        val_count (None or int): After how many examples learned will an accuracy test be performed? default: 10=(%10) it means every approximately 10 step (optional)

        activation_potentiation (list): For deeper PLAN networks, activation function parameters. For more information please run this code: plan.activations_list() default: [None] (optional)

        x_val (list[num]): List of validation data. default: x_train (optional)

        y_val (list[num]): (list[num]): List of target labels. (one hot encoded) default: y_train (optional)

        show_training (bool, str): True or None default: None (optional)

        LTD (int): Long Term Depression Hyperparameter for train PLAN neural network default: 0 (optional)

        interval (float, int): frame delay (milisecond) parameter for Training Report (show_training=True) This parameter effects to your Training Report performance. Lower value is more diffucult for Low end PC's (33.33 = 30 FPS, 16.67 = 60 FPS) default: 100 (optional)

        decision_boundary_status (bool): If the visualization of validation and training history is enabled during training, should the decision boundaries also be visualized? True or False. Default is True. (optional)

        train_bar (bool): Training loading bar? True or False. Default is True. (optional)

        auto_normalization(bool): Normalization process during training. May effect training time and model quality. True or False. Default is True. (optional)

        neurons_history (bool, optional): Shows the history of changes that neurons undergo during the CL (Cumulative Learning) stages. True or False. Default is False. (optional)

        dtype (cupy.dtype): Data type for the arrays. np.float32 by default. Example: cp.float64 or cp.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

        memory (str): The memory parameter determines whether the dataset to be processed on the GPU will be stored in the CPU's RAM or the GPU's RAM. Options:  'gpu', 'cpu'. Default: 'gpu'.

    Returns:
        numpyarray([num]): (Weight matrix).
    """
    # Pre-checks
    
    if train_bar and val:
        train_progress = initialize_loading_bar(total=len(x_train), ncols=71, desc='Fitting', bar_format=bar_format_normal)
    elif train_bar and val == False:
        train_progress = initialize_loading_bar(total=len(x_train), ncols=44, desc='Fitting', bar_format=bar_format_normal)

    if len(x_train) != len(y_train):
        raise ValueError("x_train and y_train must have the same length.")

    if val and (x_val is None or y_val is None):
        x_val, y_val = x_train, y_train

    if memory == 'gpu':
        x_train = transfer_to_gpu(x_train, dtype=dtype)
        y_train = transfer_to_gpu(y_train, dtype=y_train.dtype)

        if val:
            x_val = transfer_to_gpu(x_val, dtype=dtype)
            y_val = transfer_to_gpu(y_val, dtype=y_train.dtype)

    elif memory == 'cpu':
        x_train = transfer_to_cpu(x_train, dtype=dtype)
        y_train = transfer_to_cpu(y_train, dtype=y_train.dtype)

        if val:
            x_val = transfer_to_cpu(x_val, dtype=dtype)
            y_val = transfer_to_cpu(y_val, dtype=y_train.dtype)

    else:
        raise ValueError("memory parameter must be 'cpu' or 'gpu'.")

    val_list = [] if val else None
    val_count = val_count or 10
    # Defining weights
    STPW = cp.ones((len(y_train[0]), len(x_train[0].ravel()))).astype(dtype, copy=False)  # STPW = SHORT TERM POTENTIATION WEIGHT
    LTPW = cp.zeros((len(y_train[0]), len(x_train[0].ravel()))).astype(dtype, copy=False)  # LTPW = LONG TERM POTENTIATION WEIGHT
    # Initialize visualization
    vis_objects = initialize_visualization_for_fit(val, show_training, neurons_history, x_train, y_train)
   
    # Training process
    for index, inp in enumerate(x_train):
        inp = transfer_to_gpu(inp, dtype=dtype).ravel()
        y_decoded = decode_one_hot(cp.array(y_train[index], copy=False, dtype=y_train.dtype))
        # Weight updates
        STPW = feed_forward(inp, STPW, is_training=True, Class=y_decoded, activation_potentiation=activation_potentiation, LTD=LTD)
        LTPW += normalization(STPW, dtype=dtype) if auto_normalization else STPW

        if val and index != 0:
            if index % math.ceil((val_count / len(x_train)) * 100) == 0:
                val_acc = evaluate(x_val, y_val, loading_bar_status=False, activation_potentiation=activation_potentiation, W=LTPW, memory=memory)[get_acc()]
                val_list.append(val_acc)

                # Visualization updates
                if show_training:
                    update_weight_visualization_for_fit(vis_objects['ax'][0, 0], LTPW, vis_objects['artist2'])
                    if decision_boundary_status:
                        update_decision_boundary_for_fit(vis_objects['ax'][0, 1], x_val, y_val, activation_potentiation, LTPW, vis_objects['artist1'])
                    update_validation_history_for_fit(vis_objects['ax'][1, 1], val_list, vis_objects['artist3'])
                    update_neural_web_for_fit(W=LTPW, G=vis_objects['G'], ax=vis_objects['ax'][1, 0], artist=vis_objects['artist4'])
                if neurons_history:
                    update_neuron_history(LTPW, row=vis_objects['row'], col=vis_objects['col'], class_count=len(y_train[0]), fig1=vis_objects['fig1'], ax1=vis_objects['ax1'], artist5=vis_objects['artist5'], acc=val_acc)
        if train_bar:
             train_progress.update(1)

        STPW = cp.ones((len(y_train[0]), len(x_train[0].ravel()))).astype(dtype, copy=False)

    if show_training:
        ani1 = display_visualization_for_fit(vis_objects['fig'], vis_objects['artist1'], interval)
        ani2 = display_visualization_for_fit(vis_objects['fig'], vis_objects['artist2'], interval)
        ani3 = display_visualization_for_fit(vis_objects['fig'], vis_objects['artist3'], interval)
        ani4 = display_visualization_for_fit(vis_objects['fig'], vis_objects['artist4'], interval)
        show()
    
    if neurons_history:
        ani5 = display_visualization_for_fit(vis_objects['fig1'], vis_objects['artist5'], interval)
        show()

    return normalization(LTPW, dtype=dtype)


def learner(x_train, y_train, optimizer, x_test=None, y_test=None, strategy='accuracy', gen=None, batch_size=1,
           neural_web_history=False, show_current_activations=False, auto_normalization=True,
           neurons_history=False, early_stop=False, loss='categorical_crossentropy', show_history=False,
           interval=33.33, target_acc=None, target_loss=None, except_this=None,
           only_this=None, start_this_act=None, start_this_W=None, target_fitness='max', dtype=cp.float32, memory='gpu'):
    """
    Optimizes the activation functions for a neural network by leveraging train data to find 
    the most accurate combination of activation potentiation for the given dataset.
    
    Why genetic optimization and not backpropagation?
    Because PLAN is different from other neural network architectures. In PLAN, the learnable parameters are not the weights; instead, the learnable parameters are the activation functions.
    Since activation functions are not differentiable, we cannot use gradient descent or backpropagation. However, I developed a more powerful genetic optimization algorithm: PLANEAT.

    Args:

        x_train (array-like): Training input data.

        y_train (array-like): Labels for training data.

        optimizer (function): PLAN optimization technique with hyperparameters. (PLAN using NEAT(PLANEAT) for optimization.) Please use this: from pyerualjetwork import planeat_cuda (and) optimizer = lambda *args, **kwargs: planeat_cuda.evolve(*args, 'here give your neat hyperparameters for example:  activation_add_prob=0.85', **kwargs) Example:
        ```python
        genetic_optimizer = lambda *args, **kwargs: planeat_cuda.evolve(*args,
                                                                    activation_add_prob=0.85,
                                                                    mutations=False,
                                                                    strategy='cross_over',
                                                                    **kwargs)

        model = plan_cuda.learner(x_train,
                             y_train,
                             optimizer=genetic_optimizer,
                             strategy='accuracy',
                             show_history=True,
                             target_acc=0.94,
                             interval=16.67)
        ```

        x_test (array-like, optional): Test input data (for improve next gen generilization). If test data is not given then train feedback learning active

        y_test (array-like, optional): Test Labels (for improve next gen generilization). If test data is not given then train feedback learning active

        strategy (str, optional): Learning strategy. (options: 'accuracy', 'f1', 'precision', 'recall'): 'accuracy', Maximizes train (or test if given) accuracy during learning. 'f1', Maximizes train (or test if given) f1 score during learning. 'precision', Maximizes train (or test if given) precision score during learning. 'recall', Maximizes train (or test if given) recall during learning. Default is 'accuracy'.

        patience ((int, float), optional): patience value for adaptive strategies. For 'adaptive_accuracy' Default value: 5. For 'adaptive_loss' Default value: 0.150.

        gen (int, optional): The generation count for genetic optimization.

        batch_size (float, optional): Batch size is used in the prediction process to receive test feedback by dividing the test data into chunks and selecting activations based on randomly chosen partitions. This process reduces computational cost and time while still covering the entire test set due to random selection, so it doesn't significantly impact accuracy. For example, a batch size of 0.08 means each test batch represents 8% of the test set. Default is 1. (%100 of test)

        auto_normalization (bool, optional): If auto normalization=False this makes more faster training times and much better accuracy performance for some datasets. Default is True.

        early_stop (bool, optional): If True, implements early stopping during training.(If test accuracy not improves in two gen stops learning.) Default is False.

        show_current_activations (bool, optional): Should it display the activations selected according to the current strategies during learning, or not? (True or False) This can be very useful if you want to cancel the learning process and resume from where you left off later. After canceling, you will need to view the live training activations in order to choose the activations to be given to the 'start_this' parameter. Default is False

        show_history (bool, optional): If True, displays the training history after optimization. Default is False.

        loss (str, optional): For visualizing and monitoring. PLAN neural networks doesn't need any loss function in training. options: ('categorical_crossentropy' or 'binary_crossentropy') Default is 'categorical_crossentropy'.

        interval (int, optional): The interval at which evaluations are conducted during training. (33.33 = 30 FPS, 16.67 = 60 FPS) Default is 100.

        target_acc (int, optional): The target accuracy to stop training early when achieved. Default is None.

        target_loss (float, optional): The target loss to stop training early when achieved. Default is None.

        except_this (list, optional): A list of activations to exclude from optimization. Default is None. (For avaliable activation functions, run this code: plan.activations_list())

        only_this (list, optional): A list of activations to focus on during optimization. Default is None. (For avaliable activation functions, run this code: plan.activations_list())
        
        start_this_act (list, optional): To resume a previously canceled or interrupted training from where it left off, or to continue from that point with a different strategy, provide the list of activation functions selected up to the learned portion to this parameter. Default is None

        start_this_W (cupy.array, optional): To resume a previously canceled or interrupted training from where it left off, or to continue from that point with a different strategy, provide the weight matrix of this genome. Default is None
        
        neurons_history (bool, optional): Shows the history of changes that neurons undergo during the TFL (Test or Train Feedback Learning) stages. True or False. Default is False.

        neural_web_history (bool, optional): Draws history of neural web. Default is False.

        dtype (cupy.dtype): Data type for the arrays. np.float32 by default. Example: cp.float64 or cp.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

        memory (str): The memory parameter determines whether the dataset to be processed on the GPU will be stored in the CPU's RAM or the GPU's RAM. Options:  'gpu', 'cpu'. Default: 'gpu'.

    Returns:
        tuple: A list for model parameters: [Weight matrix, Preds, Accuracy, [Activations functions]]. You can acces this parameters in model_operations module. For example: model_operations.get_weights() for Weight matrix.
    
    """

    print(Fore.WHITE + "\nRemember, optimization on large datasets can be very time-consuming and computationally expensive. Therefore, if you are working with such a dataset, our recommendation is to include activation function: ['circular', 'spiral'] in the 'except_this' parameter unless absolutely necessary, as they can significantly prolong the process. from: learner\n" + Fore.RESET)

    activation_potentiation = all_activations()

    y_train = optimize_labels(y_train, cuda=True)

    if x_test is None and y_test is None:
        x_test = x_train
        y_test = y_train
        data = 'Train'
    else:
        data = 'Test'
        y_test = optimize_labels(y_test, cuda=True)
        
    if memory == 'gpu':
        x_train = transfer_to_gpu(x_train, dtype=dtype)
        y_train = transfer_to_gpu(y_train, dtype=y_train.dtype)

        x_test = transfer_to_gpu(x_test, dtype=dtype)
        y_test = transfer_to_gpu(y_test, dtype=y_train.dtype)

        from .data_operations_cuda import batcher

    elif memory == 'cpu':
        x_train = transfer_to_cpu(x_train, dtype=dtype)
        y_train = transfer_to_cpu(y_train, dtype=y_train.dtype)

        x_test = transfer_to_cpu(x_test, dtype=dtype)
        y_test = transfer_to_cpu(y_test, dtype=y_train.dtype)

        from .data_operations import batcher

    else:
        raise ValueError("memory parameter must be 'cpu' or 'gpu'.")

    if strategy != 'accuracy' and strategy != 'f1' and strategy != 'recall' and strategy != 'precision': raise ValueError("Strategy parameter only be 'accuracy' or 'f1' or 'recall' or 'precision'.")

    # Filter activation functions
    if only_this is not None:
        activation_potentiation = only_this
    if except_this is not None:
        activation_potentiation = [item for item in activation_potentiation if item not in except_this]
    if gen is None:
        gen = len(activation_potentiation)

    if start_this_act is None and len(activation_potentiation) % 2 != 0: raise ValueError("Activation length must be even number. Please use 'except_this' parameter and except some activation. For example: except_this=['linear']")
    if start_this_act is not None and len(activation_potentiation) + 1 % 2 != 0: raise ValueError("You are using start_this parameter, activation length still must be even number. Please use 'except_this' parameter and except some activation. For example: except_this=['linear']")
    
    # Initialize visualization components
    viz_objects = initialize_visualization_for_learner(show_history, neurons_history, neural_web_history, x_train, y_train)

    # Initialize progress bar
    if batch_size == 1:
        ncols = 86
    else:
        ncols = 99

    # Initialize variables
    act_pop = []
    weight_pop = []
    
    if start_this_act is None and start_this_W is None:
        best_acc = 0
    else:
        act_pop.append(start_this_act)
        weight_pop.append(start_this_W)
        
        x_test_batch, y_test_batch = batcher(x_test, y_test, batch_size=batch_size)
        model = evaluate(x_test_batch, y_test_batch, W=start_this_W, loading_bar_status=False, activation_potentiation=act_pop, dtype=dtype, memory=memory)

        if loss == 'categorical_crossentropy':
            test_loss = categorical_crossentropy(y_true_batch=transfer_to_gpu(y_test_batch, dtype=y_test_batch.dtype), y_pred_batch=model[get_preds_softmax()])
        else:
            test_loss = binary_crossentropy(y_true_batch=transfer_to_gpu(y_test_batch, dtype=y_test_batch.dtype), y_pred_batch=model[get_preds_softmax()])

        best_acc = model[get_acc()]
        best_loss = test_loss

    best_acc_per_gen_list = []
    postfix_dict = {}
    loss_list = []
    target_pop = []

    progress = initialize_loading_bar(total=len(activation_potentiation), desc="", ncols=ncols, bar_format=bar_format_learner)

    for i in range(gen):
        postfix_dict["Gen"] = str(i+1) + '/' + str(gen)
        progress.set_postfix(postfix_dict)

        progress.n = 0
        progress.last_print_n = 0
        progress.update(0)

        for j in range(len(activation_potentiation)):

            x_test_batch, y_test_batch = batcher(x_test, y_test, batch_size=batch_size)
            
            if i == 0:
                act_pop.append(activation_potentiation[j])
                W = fit(x_train, y_train, activation_potentiation=act_pop[-1], train_bar=False, auto_normalization=auto_normalization, dtype=dtype)
                weight_pop.append(W)

            model = evaluate(x_test_batch, y_test_batch, W=weight_pop[j], loading_bar_status=False, activation_potentiation=act_pop[j], dtype=dtype, memory=memory)

            acc = model[get_acc()]
            if strategy == 'accuracy': target_pop.append(acc)

            elif strategy == 'f1' or strategy == 'precision' or strategy == 'recall':
                precision_score, recall_score, f1_score = metrics(y_test_batch, model[get_preds()])

                if strategy == 'precision':
                    target_pop.append(precision_score)

                    if i == 0 and j == 0:
                        best_precision = precision_score

                if strategy == 'recall':
                    target_pop.append(recall_score)

                    if i == 0 and j == 0:
                        best_recall = recall_score

                if strategy == 'f1':
                    target_pop.append(f1_score)

                    if i == 0 and j == 0:
                        best_f1 = f1_score

            if ((strategy == 'accuracy' and acc >= best_acc) or 
                (strategy == 'f1' and f1_score >= best_f1) or 
                (strategy == 'precision' and precision_score >= best_precision) or 
                (strategy == 'recall' and recall_score >= best_recall)):

                best_acc = acc
                best_weights = weight_pop[j]
                final_activations = act_pop[j]
                best_model = model

                final_activations = [final_activations[0]] if len(set(final_activations)) == 1 else final_activations # removing if all same

                if batch_size == 1:
                    postfix_dict[f"{data} Accuracy"] = best_acc
                else:
                    postfix_dict[f"{data} Batch Accuracy"] = acc
                progress.set_postfix(postfix_dict)

                if show_current_activations:
                    print(f", Current Activations={final_activations}", end='')
            
                if loss == 'categorical_crossentropy':
                    test_loss = categorical_crossentropy(y_true_batch=transfer_to_gpu(y_test_batch, dtype=y_test_batch.dtype), y_pred_batch=model[get_preds_softmax()])
                else:
                    test_loss = binary_crossentropy(y_true_batch=transfer_to_gpu(y_test_batch, dtype=y_test_batch.dtype), y_pred_batch=model[get_preds_softmax()])

                if batch_size == 1:
                    postfix_dict[f"{data} Loss"] = test_loss
                    best_loss = test_loss
                else:
                    postfix_dict[f"{data} Batch Loss"] = test_loss
                progress.set_postfix(postfix_dict)
                best_loss = test_loss

                # Update visualizations during training
                if show_history:
                    gen_list = range(1, len(best_acc_per_gen_list) + 2)
                    update_history_plots_for_learner(viz_objects, gen_list, loss_list + [test_loss], 
                                      best_acc_per_gen_list + [best_acc], x_train, final_activations)

                if neurons_history:
                    viz_objects['neurons']['artists'] = (
                        update_neuron_history_for_learner(cp.copy(best_weights), viz_objects['neurons']['ax'],
                                     viz_objects['neurons']['row'], viz_objects['neurons']['col'],
                                     y_train[0], viz_objects['neurons']['artists'],
                                     data=data, fig1=viz_objects['neurons']['fig'],
                                     acc=best_acc, loss=test_loss)
                    )

                if neural_web_history:
                    art5_1, art5_2, art5_3 = draw_neural_web(W=best_weights, ax=viz_objects['web']['ax'],
                                                            G=viz_objects['web']['G'], return_objs=True)
                    art5_list = [art5_1] + [art5_2] + list(art5_3.values())
                    viz_objects['web']['artists'].append(art5_list)

                # Check target accuracy
                if target_acc is not None and best_acc >= target_acc:
                    progress.close()
                    train_model = evaluate(x_train, y_train, W=best_weights, loading_bar_status=False, 
                                        activation_potentiation=final_activations, dtype=dtype)

                    if loss == 'categorical_crossentropy':
                        train_loss = categorical_crossentropy(y_true_batch=y_train, 
                                                           y_pred_batch=train_model[get_preds_softmax()])
                    else:
                        train_loss = binary_crossentropy(y_true_batch=y_train, 
                                                       y_pred_batch=train_model[get_preds_softmax()])

                    print('\nActivations: ', final_activations)
                    print(f'Train Accuracy (%{batch_size * 100} of train samples):', train_model[get_acc()])
                    print(f'Train Loss (%{batch_size * 100} of train samples): ', train_loss, '\n')
                    if data == 'Test':
                        print(f'Test Accuracy (%{batch_size * 100} of test samples): ', best_acc)
                        print(f'Test Loss (%{batch_size * 100} of test samples): ', best_loss, '\n')

                    # Display final visualizations
                    display_visualizations_for_learner(viz_objects, best_weights, data, best_acc, 
                                              test_loss, y_train, interval)
                    return best_weights, best_model[get_preds()], best_acc, final_activations

                # Check target loss
                if target_loss is not None and best_loss <= target_loss:
                    progress.close()
                    train_model = evaluate(x_train, y_train, W=best_weights, loading_bar_status=False, 
                                        activation_potentiation=final_activations, dtype=dtype)

                    if loss == 'categorical_crossentropy':
                        train_loss = categorical_crossentropy(y_true_batch=y_train, 
                                                           y_pred_batch=train_model[get_preds_softmax()])
                    else:
                        train_loss = binary_crossentropy(y_true_batch=y_train, 
                                                       y_pred_batch=train_model[get_preds_softmax()])

                    print('\nActivations: ', final_activations)
                    print(f'Train Accuracy (%{batch_size * 100} of train samples):', train_model[get_acc()])
                    print(f'Train Loss (%{batch_size * 100} of train samples): ', train_loss, '\n')
                    if data == 'Test':
                        print(f'Test Accuracy (%{batch_size * 100} of test samples): ', best_acc)
                        print(f'Test Loss (%{batch_size * 100} of test samples): ', best_loss, '\n')

                    # Display final visualizations
                    display_visualizations_for_learner(viz_objects, best_weights, data, best_acc, 
                                              test_loss, y_train, interval)
                    return best_weights, best_model[get_preds()], best_acc, final_activations

            progress.update(1)

        best_acc_per_gen_list.append(best_acc)
        loss_list.append(best_loss)
        
        weight_pop, act_pop = optimizer(cp.array(weight_pop, dtype=dtype), act_pop, i, cp.array(target_pop, dtype=dtype, copy=False), target_fitness=target_fitness, bar_status=False)
        target_pop = []

        # Early stopping check
        if early_stop == True and i > 0:
            if best_acc_per_gen_list[i] == best_acc_per_gen_list[i-1]:
                progress.close()
                train_model = evaluate(x_train, y_train, W=best_weights, loading_bar_status=False, 
                                    activation_potentiation=final_activations, dtype=dtype, memory=memory)

                if loss == 'categorical_crossentropy':
                    train_loss = categorical_crossentropy(y_true_batch=y_train, 
                                                       y_pred_batch=train_model[get_preds_softmax()])
                else:
                    train_loss = binary_crossentropy(y_true_batch=y_train, 
                                                   y_pred_batch=train_model[get_preds_softmax()])

                print('\nActivations: ', final_activations)
                print(f'Train Accuracy (%{batch_size * 100} of train samples):', train_model[get_acc()])
                print(f'Train Loss (%{batch_size * 100} of train samples): ', train_loss, '\n')
                if data == 'Test':
                    print(f'Test Accuracy (%{batch_size * 100} of test samples): ', best_acc)
                    print(f'Test Loss (%{batch_size * 100} of test samples): ', best_loss, '\n')

                # Display final visualizations
                display_visualizations_for_learner(viz_objects, best_weights, data, best_acc, 
                                          test_loss, y_train, interval)
                return best_weights, best_model[get_preds()], best_acc, final_activations

    # Final evaluation
    progress.close()
    train_model = evaluate(x_train, y_train, W=best_weights, loading_bar_status=False, 
                        activation_potentiation=final_activations, dtype=dtype, memory=memory)

    if loss == 'categorical_crossentropy':
        train_loss = categorical_crossentropy(y_true_batch=y_train, y_pred_batch=train_model[get_preds_softmax()])
    else:
        train_loss = binary_crossentropy(y_true_batch=y_train, y_pred_batch=train_model[get_preds_softmax()])

    print('\nActivations: ', final_activations)
    print(f'Train Accuracy (%{batch_size * 100} of train samples):', train_model[get_acc()])
    print(f'Train Loss (%{batch_size * 100} of train samples): ', train_loss, '\n')
    if data == 'Test':
        print(f'Test Accuracy (%{batch_size * 100} of test samples): ', best_acc)
        print(f'Test Loss (%{batch_size * 100} of test samples): ', best_loss, '\n')

    # Display final visualizations
    display_visualizations_for_learner(viz_objects, best_weights, data, best_acc, test_loss, y_train, interval)
    return best_weights, best_model[get_preds()], best_acc, final_activations



def feed_forward(
    Input,               # list[num]: Input data.
    w,                   # num: Weight matrix of the neural network.
    is_training,        # bool: Flag indicating if the function is called during training (True or False).
    activation_potentiation,
    Class='?',               # int: Which class is, if training. # (list): Activation potentiation list for deep PLAN. (optional)
    LTD=0
) -> tuple:
    """
    Applies feature extraction process to the input data using synaptic potentiation.

    Args:
        Input (num): Input data.
        w (num): Weight matrix of the neural network.
        is_training (bool): Flag indicating if the function is called during training (True or False).
        Class (int): if is during training then which class(label) ? is isnt then put None.
        # activation_potentiation (list): ac list for deep PLAN. default: [None] ('linear') (optional)

    Returns:
        tuple: A tuple (vector) containing the neural layer result and the updated weight matrix.
        or
        num: neural network output
    """
    
    Output = apply_activation(Input, activation_potentiation)

    Input = Output

    if is_training == True:
        
        for _ in range(LTD):        

            depression_vector = cp.random.rand(*Input.shape)

            Input -= depression_vector

        w[Class, :] = Input
        return w

    else:

        neural_layer = cp.dot(w, Input)

        return neural_layer
   

def evaluate(
    x_test,
    y_test,
    W,
    activation_potentiation=['linear'],
    loading_bar_status=True,
    show_metrics=False,
    dtype=cp.float32,
    memory='gpu'
) -> tuple:
    """
    Evaluates the neural network model using the given test data.

    Args:
        x_test (cp.ndarray): Test data.

        y_test (cp.ndarray): Test labels (one-hot encoded).

        W (list[cp.ndarray]): Neural net weight matrix.
        
        activation_potentiation (list): Activation list. Default = ['linear'].
        
        loading_bar_status (bool): Loading bar (optional). Default = True.

        show_metrics (bool): Visualize metrics ? (optional). Default = False.

        dtype (cupy.dtype): Data type for the arrays. cp.float32 by default. Example: cp.float64 or cp.float16. [fp32 for balanced devices, fp64 for strong devices, fp16 for weak devices: not reccomended!] (optional)

        memory (str): The memory parameter determines whether the dataset to be processed on the GPU will be stored in the CPU's RAM or the GPU's RAM. Options:  'gpu', 'cpu'. Default: 'gpu'.

    Returns:
        tuple: Model (list).
    """
    from memory_operations import transfer_to_cpu, transfer_to_gpu

    if memory == 'gpu':
        x_test = transfer_to_gpu(x_test, dtype=dtype)
        y_test = transfer_to_gpu(y_test, dtype=y_test.dtype)

    elif memory == 'cpu':
        x_test = transfer_to_cpu(x_test, dtype=dtype)
        y_test = transfer_to_cpu(y_test, dtype=y_test.dtype)

    else:
        raise ValueError("memory parameter must be 'cpu' or 'gpu'.")

    predict_probabilitys = cp.empty((len(x_test), W.shape[0]), dtype=dtype)
    real_classes = cp.empty(len(x_test), dtype=y_test.dtype)   
    predict_classes = cp.empty(len(x_test), dtype=y_test.dtype)

    true_predict = 0
    acc_list = cp.empty(len(x_test), dtype=dtype)

    if loading_bar_status:
        loading_bar = initialize_loading_bar(total=len(x_test), ncols=64, desc='Testing', bar_format=bar_format_normal)

    for inpIndex in range(len(x_test)):
        Input = transfer_to_gpu(x_test[inpIndex], dtype=dtype).ravel()
        neural_layer = Input

        neural_layer = feed_forward(neural_layer, cp.copy(W), is_training=False, Class='?', activation_potentiation=activation_potentiation)

        predict_probabilitys[inpIndex] = Softmax(neural_layer)

        RealOutput = decode_one_hot(transfer_to_gpu(y_test[inpIndex], dtype=y_test[inpIndex].dtype))
        real_classes[inpIndex] = RealOutput
        PredictedOutput = cp.argmax(neural_layer)
        predict_classes[inpIndex] = PredictedOutput

        if RealOutput == PredictedOutput:
            true_predict += 1
        
        acc = true_predict / (inpIndex + 1)
        acc_list[inpIndex] = acc

        if loading_bar_status:
            loading_bar.update(1)
            loading_bar.set_postfix({"Test Accuracy": acc})

    if loading_bar_status:
        loading_bar.close()

    if show_metrics:
        plot_evaluate(x_test, y_test, predict_classes, acc_list, W=cp.copy(W), activation_potentiation=activation_potentiation)

    return W, predict_classes, acc_list[-1], None, None, predict_probabilitys